/* 

robotAI
Version 1
June 18 2008
Released under GNU GPL

robotAI ships freight. with trucks

*/

import("pathfinder.road", "RoadPathFinder", 3);

//
//
// Class - Line
//
//

class Line {
	source = null;
	destination = null;
	station1 = null;
	station1tile = null;
	station2 = null;
	station2tile = null;
	depottile = null;
	Vehicles = null;
	cargo = null;

	constructor(source, destination, cargo) {
		this.source = source;
		this.destination = destination;
		this.cargo = cargo;
		this.Vehicles = [];
	}

	function InitializeRoute();
	function FindSite(target, siteType, radius);
	function GetCargoRating();
	function AddAdditionalVehicle();
}

//
// function Line.InitializeRoute;
//

function Line::InitializeRoute() 
{
	// AILog.Info("Line: initializing route");	
	
	/* find choose a tile for the source and destination stations.*/
	local sourceSite = AIIndustry.GetLocation(source);
	local destinationSite = AIIndustry.GetLocation(destination);	
	local station1Site = this.FindSite(sourceSite, 1, 4);
	local station2Site = this.FindSite(destinationSite, 2, 4);
	if (station1Site[0] == -1 || station1Site[1] == 1) { 
		AILog.Info("Line: Could not locate station 1!");
		return false;	
	}
	if (station1Site[0] == -1 || station1Site[1] == 1) { 
		AILog.Info("Line: Could not locate station 2!");
		return false;	
	}

	/* build stations */

	if (!AIRoad.BuildRoad(station1Site[0], station1Site[1])) {
		AILog.Warning("Line: Error building road 1 at " + station1Site[1]);		
		return false;
	}
	if (!AIRoad.BuildRoad(station2Site[0], station2Site[1])) {
		AILog.Warning("Line: Error building road 2 at " + station2Site[1]);
		return false;
	}
	
 
	if (!AIRoad.BuildRoadStation(station1Site[0], station1Site[1], true, false)) {
		AILog.Warning("Line: Error building station 1 at " + station1Site[0]);
		return false;
	}
	if (!AIRoad.BuildRoadStation(station2Site[0], station2Site[1], true, false)) {
		AILog.Warning("Line: Error building station 2 at" + station2Site[0]);
		return false;
	}	

	// find the route	
	local pathfinder = RoadPathFinder();
	pathfinder.cost.no_existing_road = 400;
/*	pathfinder.cost.turn = 750;
	pathfinder.cost.coast = 500;
	pathfinder.cost.slope = 500;
	pathfinder.cost.max_tunnel_length = 30;
	pathfinder.cost.max_bridge_length = 30; */
	pathfinder.cost.bridge_per_tile = 1500;
	pathfinder.cost.tunnel_per_tile = 1300;


	pathfinder.InitializePath([station1Site[1]], [station2Site[1]]);

	local path = false;
	while (path == false) {
		path = pathfinder.FindPath(100);
		// AILog.Info("Line: Finding route...");		
		robotAI.Sleep(1);
	}

	if (path == null) {
		AILog.Info("Line: no path found!");
		return false;
	}

	AILog.Info("Line: Found route");
	while (path != null) {
		local par = path.GetParent();
		if (path.GetParent() != null) {
			
			if (AIMap.DistanceManhattan(path.GetTile(), par.GetTile()) == 1 ) {

// it is possible part of the path is blocked, so if building a tile fails
// check for that
				// AILog.Info("Line: Building road...");
				while (!AIRoad.BuildRoad(path.GetTile(), par.GetTile())) {
					if (AIError.GetLastError() == AIError.ERR_VEHICLE_IN_THE_WAY) {
						AILog.Info("Line: Road vehicle in the way, retrying.");
						AIController.Sleep(1);
					}
					else if (AIError.GetLastError() == AIError.ERR_NOT_ENOUGH_CASH) {
						AILog.Warning("Line: Error: Not enough cash to finish building route.");
						return false;
					}
// if that's not the case, maybe the tiles are already connected
					else if (AIRoad.AreRoadTilesConnected(path.GetTile(), par.GetTile())) {
						break;
					}
// if that's not the case, whine
					else {					
						AILog.Warning("Line: Error building route");
						return false;
					}
				}
			
			} else { // it's a bridge or a tunnel
			 if (!AIBridge.IsBridgeTile(path.GetTile()) && 
					!AITunnel.IsTunnelTile(path.GetTile())) {
					if (AIRoad.IsRoadTile(path.GetTile())) AITile.DemolishTile(path.GetTile());
					if (AITunnel.GetOtherTunnelEnd(path.GetTile()) == par.GetTile()) {
						AILog.Info("Line: Building tunnel...");						
						while (!AITunnel.BuildTunnel(AIVehicle.VEHICLE_ROAD, path.GetTile())) {
							if (AIError.GetLastError() == AIError.ERR_VEHICLE_IN_THE_WAY) {
								AILog.Info("Line: Road vehicle in the way, retrying.");
								AIController.Sleep(1);
							} else if (AIError.GetLastError() == AIError.ERR_NOT_ENOUGH_CASH) {
								AILog.Warning("Line: Error: Not enough cash to finish building route.");
								return false;
							} else {
								AILog.Warning("Line: Error building route");
								return false;
							}
						}
					} else {
						local bridges = AIBridgeList_Length(AIMap.DistanceManhattan
							(path.GetTile(), par.GetTile()));
						bridges.Valuate(AIBridge.GetMaxSpeed);
						bridges.Sort(AIAbstractList.SORT_BY_VALUE, false);
						AILog.Info("Line: building bridge...");						
						while (!AIBridge.BuildBridge(AIVehicle.VEHICLE_ROAD, bridges.Begin(), 
							path.GetTile(), par.GetTile())) {
							if (AIError.GetLastError() == AIError.ERR_VEHICLE_IN_THE_WAY) {
								AILog.Info("Line: Road vehicle in the way, retrying.");
								AIController.Sleep(1);
							} else if (AIError.GetLastError() == AIError.ERR_NOT_ENOUGH_CASH) {
								AILog.Warning("Line: Error: Not enough cash to finish building route.");
								return false;
							} else {
								AILog.Warning("Line: Error building route");
								return false;
							}
						}
					}
				}
			}
		}
		path = par;
	}

	// now the path is built, we locate and build a depot
	
	AILog.Info("Line: Stations connected, building depot");
	local depotSite = this.FindSite(station1Site[0], 0, 10);
	while (!AIRoad.BuildRoad(depotSite[0], depotSite[1])) {
		if (AIError.GetLastError() == AIError.ERR_VEHICLE_IN_THE_WAY) {
			AILog.Info("Line: Road vehicle in the way, retrying.");
			AIController.Sleep(1);
		} else {
			AILog.Warning("Line: Error build road at" + depotSite[0]);
			return false;
		}
	}	
	while (!AIRoad.BuildRoadDepot(depotSite[0], depotSite[1])) {
		if (AIError.GetLastError() == AIError.ERR_VEHICLE_IN_THE_WAY) {
			AILog.Info("Line: Road vehicle in the way, retrying.");
			AIController.Sleep(1);
		} else {
			AILog.Warning("Line: Error build road at" + depotSite[0]);
			return false;
		}
	}	
	
	this.depottile = depotSite[0];
	this.station1tile = station1Site[0];
	this.station2tile = station2Site[0];
	this.station1 = AIStation.GetStationID(station1Site[0]);
	this.station2 = AIStation.GetStationID(station2Site[0]);

	// now there's two stations and a depot, so we can actually add a truck
	// find the best engine to use
	local possibleEngines = AIEngineList(AIVehicle.VEHICLE_ROAD);
	possibleEngines.Valuate(AIEngine.GetCargoType);
	possibleEngines.KeepValue(this.cargo);
	possibleEngines.Valuate(AIEngine.GetReliability);
	possibleEngines.Sort(AIList.SORT_BY_VALUE, false);
	possibleEngines.KeepTop(1);
	local engineID = possibleEngines.Begin();
	// build the vehicle
	local thisvehicle = AIVehicle.BuildVehicle(this.depottile, engineID);
	if (!AIVehicle.IsValidVehicle(thisvehicle)) {
		AILog.Warning("Line: Error building vehicle at " + this.depottile);
	}
	// give it some orders
	if (!AIOrder.AppendOrder(thisvehicle, this.station1tile,
		AIOrder.AIOF_FULL_LOAD) || !AIOrder.AppendOrder(thisvehicle, this.depottile, 
		AIOrder.AIOF_NONE) || !AIOrder.AppendOrder(thisvehicle,
		this.station2tile, AIOrder.AIOF_NONE)) {
		AILog.Warning("Line: error giving orders to vehicle " + thisvehicle);
		return false;
	}
	//
	if (!AIVehicle.StartStopVehicle(thisvehicle)) {
		AILog.Warning("Line: error starting vehicle " + thisvehicle);
		return false;
	}
	this.Vehicles.append(thisvehicle);	
	AILog.Info("Line: Line successfully built from " + 
	 	AIStation.GetName(this.station1) + " to " + 
	 	AIStation.GetName(this.station2));
	return true;

}



//
// function Line.FindSite(target, siteType, radius)
//
//
// siteType (int) - 0 depot, 1 producing station, 2 accepting station
// radius (int) - the number of tile out from the target for the search 
// 	rectangle, ie 3 searches a 7x7 grid around the target 
// 	(this is not actually what radius means)
//	
// returns [site, front]
// [-1, -1] means no suitable site was found
//

function Line::FindSite(target, siteType, radius)
{	
	local possibilities = AITileList();
	local adjacent = AITileList();
	local attempt = 0;
	local count = 0;
	local foundtile = -1; 	
	while (count == 0) {
		AIController.Sleep(1);
		attempt = attempt + 1; // this is not real problem solving
		if (attempt == 6) {
			AILog.Warning("Line: Unable to locate site!");
			return [-1,-1];		
		}		
		// AILog.Info("Line: Attempt " + attempt);
		local area = AITileList();
		area.AddRectangle(target - AIMap.GetTileIndex(radius, radius), target + 
			AIMap.GetTileIndex(radius, radius));
		area.Valuate(AIMap.IsValidTile);
		area.KeepValue(1);	
	
		if (siteType >> 0) {	// if it's a station we want empty tiles
			area.Valuate(AITile.IsBuildable);
			area.KeepValue(1);
			local coverage = AIStation.GetCoverageRadius(AIStation.STATION_TRUCK_STOP);
			if (siteType == 1) {			
				// for a producing station we want sites that have cargo
				area.Valuate(AITile.GetCargoProduction, this.cargo, 1, 1, coverage);
				area.RemoveValue(0);
			}
			else if (siteType == 2) {
				// for a recieving station we want sites that take cargo
				area.Valuate(AITile.GetCargoAcceptance, this.cargo, 1, 1, coverage);
				area.RemoveValue(0);
			}
		}
		else if (siteType == 0) {
			area.Valuate(AIRoad.IsRoadTile);	// if it's a depot we want road tiles
			area.KeepValue(1);
		}
		else {
			AILog.Error("Line: Must specify depot or station!");
			return [-1. -1];
		}
	
		area.Valuate(AITile.GetSlope);
		area.KeepValue(0);
		// AILog.Info("Line: " + area.Count() + " tiles found."); 	
	
		if (area.Count() == 0) {	
			return [-1, -1];
		}	
		
		if (siteType >> 0) { // for stations, choose a random tile
			area.Valuate(AIBase.RandItem);
		}
		else { // for depots, choose a random tile no farther than specified distance
			area.Valuate(AITile.GetDistanceManhattanToTile, target);
			area.KeepBelowValue(radius + 1);
			area.Valuate(AIBase.RandItem);
		}
		area.Sort(AIList.SORT_BY_VALUE, true);
		
		foundtile = area.Begin();

		if (siteType == 0) {
			AILog.Info("Line: " + area.Count() + " possible tiles for depot to face.");
			local connector = RoadPathFinder();
			connector.cost.no_existing_road = connector.cost.max_cost;
			connector.cost.bridge_per_tile = connector.cost.max_cost;
			connector.cost.tunnel_per_tile = connector.cost.max_cost;
			local connected = false;
			while (area.HasNext()) {
				// Need to make sure this is connected to the target		
				AILog.Info("Line: verifying connection for tile " + foundtile);				
				connector.InitializePath([foundtile], [target]);
				connected = connector.FindPath(100);

				// checking of adjacent tiles is here to not get tricked by towns
				// this probably means there's no need to check again later
				local height = AITile.GetHeight(foundtile);
				adjacent = AITileList();
				adjacent.AddTile(foundtile - AIMap.GetTileIndex(1,0));
				adjacent.AddTile(foundtile - AIMap.GetTileIndex(0,1));
				adjacent.AddTile(foundtile - AIMap.GetTileIndex(-1,0));
				adjacent.AddTile(foundtile - AIMap.GetTileIndex(0,-1));

				adjacent.Valuate(AIMap.IsValidTile);
				adjacent.KeepValue(1);

				adjacent.Valuate(AITile.GetHeight);
				adjacent.KeepValue(height);
				adjacent.Valuate(AITile.GetSlope);
				adjacent.KeepValue(0);
				adjacent.Valuate(AITile.IsBuildable);
				adjacent.KeepValue(1);

				if (adjacent.Count() == 0) {
					connected = false;
				}

				if (connected) {
					break;
				}				

				foundtile = area.Next();
				AIController.Sleep(1);
			}
			if (connected == false) {
				AILog.Info("Line: chosen site does not connect to target");
				return [-1, -1];
			}
		}		

		// do we need to do this again for depots? probably not, but for now that's
		// how it is
		local height = AITile.GetHeight(foundtile);
		adjacent = AITileList();
		adjacent.AddTile(foundtile - AIMap.GetTileIndex(1,0));
		adjacent.AddTile(foundtile - AIMap.GetTileIndex(0,1));
		adjacent.AddTile(foundtile - AIMap.GetTileIndex(-1,0));
		adjacent.AddTile(foundtile - AIMap.GetTileIndex(0,-1));
		
		adjacent.Valuate(AIMap.IsValidTile);
		adjacent.KeepValue(1);
		
		adjacent.Valuate(AITile.GetHeight);
		adjacent.KeepValue(height);
		adjacent.Valuate(AITile.GetSlope);
		adjacent.KeepValue(0);
		adjacent.Valuate(AITile.IsBuildable);
		adjacent.KeepValue(1);
		
		/* AITile.IsBuildable returns 0 for tiles with roads
			this is not appropriate behaviour for stations
			to compensate, this splits adjacent into two lists, run IsBuildable on one,
			checks the other for existing road tiles. drops 0 from each and merges
		
			for a depot we only want open tiles, so this does not need to be done */
			
		if (siteType >> 0) {
			local adjacent2 = AIList();
			adjacent2.AddList(adjacent); 
			adjacent2.Valuate(AIRoad.IsRoadTile);
			adjacent2.KeepValue(1);	
			adjacent.AddList(adjacent2);
		}
		count = adjacent.Count();
		// AILog.Info("Line: " + adjacent.Count() + " sites.");

	
	}
	
	adjacent.Valuate(AIBase.RandItem);
	adjacent.Sort(AIList.SORT_BY_VALUE, true);
	adjacent.KeepTop(1);
	local foundtile2 = adjacent.Begin();
	if (siteType >> 0) {
		return [foundtile, foundtile2];
	}
	else {
		return [foundtile2, foundtile];
	}
}

//
// function - Line.GetCargoRating()
// a wrapper for AIStation.GetCargoRating
//
// returns cargo rating at station1
//

function Line::GetCargoRating()
{
	local rating = AIStation.GetCargoRating(this.station1, this.cargo);
	return rating;
}

//
// function - Line.AddAdditionalVehicle()
// adds a new vehicle to the line and shares orders with the first
//	
// returns true if vehicle was added successfully
//

function Line::AddAdditionalVehicle()
{
	local possibleEngines = AIEngineList(AIVehicle.VEHICLE_ROAD);
	possibleEngines.Valuate(AIEngine.GetCargoType);
	possibleEngines.KeepValue(this.cargo);
	possibleEngines.Valuate(AIEngine.GetReliability);
	possibleEngines.Sort(AIList.SORT_BY_VALUE, false);
	possibleEngines.KeepTop(1);
	local engineID = possibleEngines.Begin();
	// build the vehicle
	local thisvehicle = AIVehicle.BuildVehicle(this.depottile, engineID);
	if (!AIVehicle.IsValidVehicle(thisvehicle)) {
		AILog.Warning("Line: Error building vehicle at " + this.depottile);
		return false;
	}
	// give it some orders
	if (!AIOrder.ShareOrders(thisvehicle, this.Vehicles[0])) {
		AILog.Warning("Line: error giving orders to vehicle " + thisvehicle);
		return false;
	}
	//
	if (!AIVehicle.StartStopVehicle(thisvehicle)) {
		AILog.Warning("Line: error starting vehicle " + thisvehicle);
		return false;
	}
	this.Vehicles.append(thisvehicle);	
	return true;
}


//
//
// Class - robotAI
//
//

class robotAI extends AIController {
	stop = false;
	name = null;
	myCargo = null;
	lines = [];
	cargoExclude = [];

	buildCash = 60000;   // cash on hand required for new route
	updateCash = 20000;  //   "      "      "         new vehicle
	ratingtarget = 55;   // station rating to maintain
	                     // this is a tradeoff between efficient service
	                     // and a truck conga line

	function Start();
	function SetCompanyName();
	function ChooseBestCargo(exclude);
	function HasMoney(CashNeeded);
}

//
// function RobotAI.Start
//

function robotAI::Start()
{
	/* Say hi */
	AILog.Info("Welcome to robotAI.")
	this.Sleep(1);
	

	this.SetCompanyName();
	this.name = AICompany.GetCompanyName(AICompany.MY_COMPANY);
	AILog.Info("INIT: Name set: " + this.name);
	this.myCargo = this.ChooseBestCargo(cargoExclude);	
	AILog.Info("INIT: Gonna go for cargo #" + myCargo);
	/* initialize main loop */
	/* set timing counters  */
	local timer = 0	

	local timerBuildRoute = timer;
	local timerCheckRoute = timer + 3;
	/* build lists of industries producing our cargo */
	local Producers = AIIndustryList_CargoProducing(myCargo);
	AILog.Info("INIT: " + Producers.Count() + " producer(s) in the world");
	local ProducersNoservice = AIList();
	ProducersNoservice.AddList(Producers);
	local ProducersBest = AIList();
	local ChosenProducer = null;
	local ChosenProducerProd = null;
	local ChosenProducerLocation = null;


	/* build lists of industries accepting our cargo */
	local Buyers = AIIndustryList_CargoAccepting(myCargo);
	local ChosenBuyer = 0;
	local ChosenBuyerLocation = null;
	AILog.Info("INIT: " + Buyers.Count() + " buyer(s) in the world");

	AILog.Info("INIT: Ready to rock...")

	while (!this.stop) {	/* main loop */
		

		while (timer >= timerBuildRoute) {	/* time to build a route! */

			timerBuildRoute = timer + 5;
			if (ProducersNoservice.IsEmpty()) { // handle being out of producers
				AILog.Info("INFO: No new producers to service");
				AILog.Info("INFO: Choosing new cargo");
				if (cargoExclude.len() != 0) {				
					if (cargoExclude.top() != myCargo) cargoExclude.append(myCargo);
				} else {
					cargoExclude.append(myCargo);
				}
				local newCargo = this.ChooseBestCargo(cargoExclude);
				if (newCargo != null) {
					this.myCargo = newCargo;						
					AILog.Info("INFO: New cargo is #" + myCargo);				
					Producers = AIIndustryList_CargoProducing(myCargo);
					AILog.Info("INFO: " + Producers.Count() + " producer(s) in the world");
					ProducersNoservice = AIList();
					ProducersNoservice.AddList(Producers);
					ProducersBest = AIList();
					ChosenProducer = null;
					ChosenProducerProd = null;
					ChosenProducerLocation = null;

			 	Buyers = AIIndustryList_CargoAccepting(myCargo);
					local ChosenBuyer = 0;
					local ChosenBuyerLocation = null;
					AILog.Info("INFO: " + Buyers.Count() + " buyer(s) in the world");
				} else {
					AILog.Error("INFO: Cound not choose new cargo!");
				}
			}
			else {																									/* let's get building */
				if (!this.HasMoney(buildCash)) {	/* if we have the cash */
					AILog.Info("INFO: Not enough cash to add new route");
				}
				else {
					/* first choose the 5 most productive producers and randomly
						choose one of them */
					// ProducersNoservice.Valuate(AIIndustry.GetProduction, myCargo);				
					// ProducersNoservice.Sort(AIList.SORT_BY_VALUE, false);

					ProducersBest = AIList();
					ProducersBest.AddList(ProducersNoservice);	
					ProducersBest.Valuate(AIIndustry.GetProduction, myCargo);
					ProducersBest.Sort(AIList.SORT_BY_VALUE, false);
					ProducersBest.KeepTop(5);

					ProducersBest.Valuate(AIBase.RandItem);
					ProducersBest.Sort(AIList.SORT_BY_VALUE, false);
				
					ChosenProducer = ProducersBest.Begin();	
					ChosenProducerProd = AIIndustry.GetProduction(ChosenProducer, myCargo);

					if (ChosenProducerProd == 0) {	
						/* make sure we don't get tricked by dead producers! */
						AILog.Info("INFO: Production at " +  AIIndustry.GetName(ChosenProducer)  
							+ " is 0, aborting.");
						ProducersBest.KeepTop(1);
						ProducersNoservice.RemoveList(ProducersBest);
						break;
					}

					ChosenProducerLocation = AIIndustry.GetLocation(ChosenProducer);
					AILog.Info("INFO: Building a route from " + 
							AIIndustry.GetName(ChosenProducer) + " (" + ChosenProducerProd + 
							" units produced)... ");

					/* then choose the closest buyer */
					Buyers.Valuate(AIIndustry.GetDistanceManhattanToTile, 
						ChosenProducerLocation);
					Buyers.Sort(AIList.SORT_BY_VALUE, true);
					if (Buyers.IsEmpty()) {
						AILog.Error("INFO: No buyers for this cargo, removing from list.");
						ProducersNoservice = AIList();
						break;
					}					
					ChosenBuyer = Buyers.Begin();
					ChosenBuyerLocation = AIIndustry.GetLocation(ChosenBuyer);
					AILog.Info("INFO: ... to " +  AIIndustry.GetName(ChosenBuyer) + 
							", distance: " + Buyers.GetValue(ChosenBuyer));

					if (!AIIndustry.IsCargoAccepted(ChosenBuyer, myCargo)) {
						AILog.Info("INFO: Illegitimate destination, aborting.");
						break;
					}

					/* here is where we actually build the route ... */
				
					local line = null;
					line = Line(ChosenProducer, ChosenBuyer, myCargo);
				
					if (!line.InitializeRoute()) {
						AILog.Warning("INFO: Error initializing route!");
					}
					else {
						AILog.Info("INFO: Route added successfully, adding to list...");
						lines.append(line);
					}

					/* remove this producer from the list of unserviced
						 producers */
					ProducersBest.KeepTop(1);
					ProducersNoservice.RemoveList(ProducersBest);
				}
				
			
		}		
	}
		this.Sleep(1);

		while (timer >= timerCheckRoute)	{	
			/* see if any of our routes need some sprucing up */
			timerCheckRoute = timer + 1;				
			if (this.HasMoney(updateCash)) {
				// AILog.Info("INFO: Checking if any routes need better service");
			
				local poorLines = AIList();
				foreach(index, thisline in lines) {
					local thisrating = thisline.GetCargoRating();
					// AILog.Info("INFO: Line " + index + " has a cargo rating of " + 
					// 	thisrating);
					if (thisrating <= this.ratingtarget) {
						poorLines.AddItem(index, thisrating);
						// AILog.Info("INFO: Line " + index + " needs a new truck.");
					}				
				}
				if (poorLines.Count() >> 0) {
					AILog.Info("INFO: " + poorLines.Count() + " lines need more trucks.");
					poorLines.Sort(AIList.SORT_BY_VALUE, true);
					local thisline = poorLines.Begin();			
					while (this.HasMoney(updateCash) && poorLines.HasNext()) {
						if (!lines[thisline].AddAdditionalVehicle()) {
							AILog.Warning("INFO: Could not add new truck to line " + thisline);
						}
						else {
							AILog.Info("INFO: added new truck to line " + thisline);
						}
						this.Sleep(1);
						thisline = poorLines.Next();
					}
				}
			}
			
		}
		timer = timer + 1;
		this.Sleep(40 + AIBase.RandRange(11));
	}

}

//
// function robotAI.SetCompanyName
//

function robotAI::SetCompanyName()
{
	if ( !AICompany.SetCompanyName("robotAI")) {
		local i = 2
		while (!AICompany.SetCompanyName("robotAI #" + i)) {
			this.Sleep(1);			
			i = i + 1;
		}
	}
}

//
// function robotAI.ChooseBestCargo(exclude)
//

function robotAI::ChooseBestCargo(exclude)
{
	/* this function chooses the best cargo to use from a given list
		a few ways to do this, i guess
		1) total production in the world
		2) number of sources
		3) best productivity at best source
		4) nearest recipient
		5) profit (get income, using distance of nearest recipient and some 
				estimate of time)
		6) randomly 

		originally used #2, now kinda uses #5 */

	/* identify freight...  */
	local CargoList = AICargoList();
	CargoList.Valuate(AICargo.IsFreight);
	CargoList.KeepValue(1);

	/* at least for now we don't want things that go to towns. */

	CargoList.Valuate(AICargo.GetTownEffect);
	CargoList.KeepValue(AICargo.TE_NONE);

	/* ... and let us know which one they are ... */
	
	AILog.Info("ChooseBestCargo: " + CargoList.Count() + 
			" cargos present");	
	local ExcludedCargos = AIList();
	foreach (index, thisExcluded in exclude) {
		if (CargoList.HasItem(thisExcluded)) {
			AILog.Info("ChooseBestCargo: Cargo #" + thisExcluded + " is excluded.");
			ExcludedCargos.AddItem(thisExcluded, 1);
		}
	}
	CargoList.RemoveList(ExcludedCargos);	
	
	if (CargoList.IsEmpty()) {
		AILog.Error("ChooseBestCargo: Unable to choose a cargo!");	
		return null;
	}

	local Cargo = CargoList.Begin();
	while (CargoList.HasNext()) {
		AILog.Info("ChooseBestCargo: Cargo " + Cargo + ": " + 
			AICargo.GetCargoLabel(Cargo)); 
		Cargo = CargoList.Next();
	} 


/* 1. Choose the cargo with the most sources

	// set up some variables
	local TempCargo = AIList();
	local ThisCargo = CargoList.Begin();	
	local ThisCargoNum = null;
	local BestCargo = null;	
	local BestCargoNum = null;	

	while (CargoList.HasNext()) {
		TempCargo = AIIndustryList_CargoProducing(ThisCargo);
		TempCargo.Valuate(AIIndustry.GetProduction, ThisCargo);
		TempCargo.RemoveValue(0);	// no nonproducing industries please
		ThisCargoNum = TempCargo.Count();
		// AILog.Info("ChooseBestCargo: Cargo #" + ThisCargo + " has " + 
		//		ThisCargoNum  + " producers.");

		if (ThisCargoNum > BestCargoNum) {	
			// compare this cargo to the previous best
			BestCargo = ThisCargo;
			BestCargoNum = ThisCargoNum;
		}

*/

/* 2. Most profitable over a given distance */
	local BestCargo = null;
	CargoList.Valuate(AICargo.GetCargoIncome, 75, 30);
		// these values (75 and 30) are just guesses... at some point it might not
		// be a bad idea to fine tune them and/or have a mechanism for choosing
		// appropriate values
	CargoList.Sort(AIList.SORT_BY_VALUE, false);
	BestCargo = CargoList.Begin();


	return BestCargo;	
}

//
// function robotAI.HasMoney(CashNeeded)
//

function robotAI::HasMoney(CashNeeded)
{
	local CashOnHand = AICompany.GetBankBalance(AICompany.MY_COMPANY);
	// AILog.Info("HasMoney: " + CashOnHand);
	if (CashOnHand > CashNeeded) {
		return true;
	}
	AICompany.SetLoanAmount(AICompany.GetMaxLoanAmount());
	CashOnHand = AICompany.GetBankBalance(AICompany.MY_COMPANY);
	if (CashOnHand > CashNeeded) {
		return true;
	}
	else {
		// AILog.Info("HasMoney: not enough cash");
		return false;
	}
}
